// Dr Benoit Chaperot 2025, inspired by AI Junkie, Mat Buckland 2002

//  Desc: Genetic algorithm class.This is based for manipulating std::vectors
//  of *real* numbers. Used to adjust the weights in a feedforward neural
//  network.

using System;
using System.Collections.Generic;
using UnityEngine;
using Random = UnityEngine.Random;

namespace OkayAI.GA
{
	[Serializable]
	public partial class GenAlg
	{
		void Crossover(List<double> mum,
			List<double> dad,
			List<double> baby1,
			List<double> baby2)
		{
			baby1.Clear();
			baby2.Clear();

			// just return parents as offspring dependent on the rate
			// or if parents are the same
			if ((Random.value > parameters.crossoverRate) || (mum == dad))
			{
				baby1.AddRange(mum);
				baby2.AddRange(dad);
				return;
			}

			// determine a crossover point
			int cp = Random.Range(0, parameters.chromoLength);

			// create the offspring
			baby1.AddRange(mum.GetRange(0, cp));
			baby2.AddRange(dad.GetRange(0, cp));
			baby1.AddRange(dad.GetRange(cp, parameters.chromoLength - cp));
			baby2.AddRange(mum.GetRange(cp, parameters.chromoLength - cp));
		}

		void Mutate(List<double> chromo)
		{
			// traverse the chromosome and mutate each weight dependent
			// on the mutation rate
			for (int i = 0; i < parameters.chromoLength; i++)
			{
				// do we perturb this weight?
				if (Random.value < parameters.mutationRate)
				{
					// add or subtract a small value to the weight
					chromo[i] += (Random.Range(-1f, 1f) * parameters.mutationPerturbation);
				}
			}
		}

		Genome GetChromoRoulette(List<Genome> inPopulation)
		{
			// generate a random number between 0 & total fitness count
			double Slice = Random.value * stats.totalFitness;

			// go through the chromosomes adding up the fitness so far
			double fitnessSoFar = 0;

			for (int i = 0; i < inPopulation.Count; i++)
			{
				fitnessSoFar += inPopulation[i].fitness;

				// if the fitness so far > random number return the chromo at
				// this point
				if (fitnessSoFar >= Slice)
				{
					return inPopulation[i];
				}

			}

			Debug.LogError("GetChromoRoulette() failed");
			return new Genome();
		}

		// use to introduce elitism
		void GrabElite(List<Genome> inPopulation, List<Genome> outPopulation)
		{
			// fittest genomes. Make sure we add an EVEN number or the roulette
			// wheel sampling will crash
			if (parameters.eliteNumBest * parameters.eliteNumCopies % 2 == 0)
			{
				// add the required amount of copies of the n fittest
				// to the supplied vector
				for (int j = 0; j < parameters.eliteNumBest; j++)
				{
					for (int i = 0; i < parameters.eliteNumCopies; i++)
					{
						outPopulation.Add(inPopulation[(parameters.populationSize - 1) - j]);
					}
				}
			}
		}

		public GenAlg(Parameters inParameters)
		{
			parameters = inParameters;
		}

		// this runs the GA for one generation.
		public List<Genome> Epoch()
		{
			stats.Update(CurrentPopulation);
			UpdateCurves();

			CurrentPopulation.Sort();
			NextPopulation.Clear();

			GrabElite(CurrentPopulation, NextPopulation);

			//repeat until a new population is generated
			while (NextPopulation.Count <parameters.populationSize)
			{
				// grab two chromosomes
				Genome mum = GetChromoRoulette(CurrentPopulation);
				Genome dad = GetChromoRoulette(CurrentPopulation);

				// create some offspring via crossover
				List<double> baby1 = new List<double>();
				List<double> baby2 = new List<double>();

				Crossover(mum.GetWeights(), dad.GetWeights(), baby1, baby2);

				//now we mutate
				Mutate(baby1);
				Mutate(baby2);

				//now copy into vecNewPop population
				NextPopulation.Add(new Genome(baby1, 0));
				NextPopulation.Add(new Genome(baby2, 0));
			}

			List<Genome> nextPopulation = NextPopulation;
			stats.Tick();
			return nextPopulation;
		}
	}
}
