// Dr Benoit Chaperot 2025, inspired by AI Junkie, Mat Buckland 2002

using System.Collections.Generic;
using OkayAI.GA;
using UnityEngine;
using Random = UnityEngine.Random;

public partial class Controller : MonoBehaviour
{
	public const float NumUpdatesPerSecond = 60f;

	public float fieldWidth = 400f;
	public float fieldHeight = 400f;

	public int numSweepers = 30;
	public int numMines = 40;
	float currentTime = 0f;

	public GameObject sweeperPrefab;
	public GameObject minePrefab;

	public Vector3 FieldWidth3 => new Vector3(fieldWidth, 0f, 0f);
	public Vector3 FieldHeight3 => new Vector3(0f, fieldHeight, 0f);

	public Vector3 RandomPosition => new Vector3(Random.Range(0f, fieldWidth), Random.Range(0f, fieldWidth), 0f);
	Quaternion RandomRotation => Quaternion.Euler(0f,0f,Random.Range(0f, 360f));

	readonly List<SmartSweeper> sweepers = new List<SmartSweeper>();
	readonly List<Mine> mines = new List<Mine>();
	readonly List<Vector2> minePositions = new List<Vector2>();

	public GenAlg.Parameters gaParameters = new GenAlg.Parameters();
	public GenAlg ga;

	public void DestroyAllChildren()
	{
		List<Transform> transformsToDestroy = new List<Transform>();

		for (int i = 0; i < transform.childCount; i++)
		{
			transformsToDestroy.Add(transform.GetChild(i));
		}

		foreach (Transform child in transformsToDestroy)
		{
			Destroy(child.gameObject);
		}
	}

	void Start()
	{
		DestroyAllChildren();

		for (int i=0;i<numMines;i++)
		{
			GameObject mineGO = Instantiate(minePrefab,RandomPosition,Quaternion.identity,transform);
			Mine mine = mineGO.GetComponent<Mine>();
			mine.controller = this;
			mines.Add(mine);
		}

		if (numSweepers <= 0)
		{
			return;
		}

		for (int i=0;i<numSweepers;i++)
		{
			GameObject sweeperGO = Instantiate(sweeperPrefab,RandomPosition,RandomRotation,transform);
			SmartSweeper sweeper = sweeperGO.GetComponent<SmartSweeper>();
			sweeper.controller = this;
			sweepers.Add(sweeper);
		}

		gaParameters.populationSize = numSweepers;
		gaParameters.chromoLength = sweepers[0].NumWeights;
		ga = new GenAlg(gaParameters);

		if (ga.Load())
		{
			gaParameters = ga.parameters;
			List<Genome> newGenomes = ga.CurrentPopulation;

			for (int i = 0; i < numSweepers; i++)
			{
				sweepers[i].PutWeights(newGenomes[i].weights);
			}
		}
		else
		{
			ga.CurrentPopulation.Clear();

			for (int i = 0; i < numSweepers; i++)
			{
				Genome genome = new Genome(0.0);
				sweepers[i].GetWeights(genome.weights);
				ga.CurrentPopulation.Add(genome);
			}
		}

		UpdateDisplay();
	}

	void OnValidate()
	{
		display.OnValidate(fieldWidth, fieldHeight);
	}

	void UpdateMinePositions()
	{
		minePositions.Clear();
		foreach (Mine mine in mines)
		{
			minePositions.Add(mine.transform.position);
		}
	}

	void FixedUpdate()
	{
		if (currentTime > gaParameters.generationTime)
		{
			for (int i = 0; i < numSweepers; i++)
			{
				Genome genome = ga.CurrentPopulation[i];
				genome.fitness = sweepers[i].score;
				ga.CurrentPopulation[i] = genome;
			}

			List<Genome> newGenomes = ga.Epoch();
			ga.Save();
			UpdateDisplay();

			for (int i = 0; i < numSweepers; i++)
			{
				sweepers[i].PutWeights(newGenomes[i].weights);
			}

			currentTime = 0f;
		}

		currentTime += Time.fixedDeltaTime;

		UpdateMinePositions();

		foreach (SmartSweeper sweeper in sweepers)
		{
			sweeper.UpdateSweeper(minePositions);
		}
	}
}
